use enum_ordinalize::Ordinalize;
use enumset::{enum_set, EnumSet, EnumSetType};

#[derive(Debug, Ordinalize, EnumSetType)]
#[repr(u16)]
pub enum LoupedeckDeviceKnob {
    KnobTopLeft = 0x01,
    KnobCenterLeft = 0x02,
    KnobBottomLeft = 0x03,
    KnobTopRight = 0x04,
    KnobCenterRight = 0x05,
    KnobBottomRight = 0x06,
}

#[derive(Debug, Ordinalize, EnumSetType)]
#[repr(u16)]
pub enum LoupedeckDeviceButton {
    KnobTopLeft = 0x01,
    KnobCenterLeft = 0x02,
    KnobBottomLeft = 0x03,
    KnobTopRight = 0x04,
    KnobCenterRight = 0x05,
    KnobBottomRight = 0x06,
    N0 = 0x07,
    N1 = 0x08,
    N2 = 0x09,
    N3 = 0x0a,
    N4 = 0x0b,
    N5 = 0x0c,
    N6 = 0x0d,
    N7 = 0x0e,
}

#[derive(Debug)]
pub enum LoupedeckDeviceDisplayEndianness {
    BigEndian,
    LittleEndian
}

#[derive(Debug)]
pub struct LoupedeckDeviceDisplayConfiguration {
    pub id: u8,
    pub width: usize,
    pub height: usize,
    pub offset: (usize, usize),
    pub endianness: LoupedeckDeviceDisplayEndianness
}

#[derive(Debug)]
pub struct LoupedeckDeviceCharacteristics {
    pub vendor_id: u16,
    pub product_id: u16,
    pub name: &'static str,
    pub available_knobs: EnumSet<LoupedeckDeviceKnob>,
    pub available_buttons: EnumSet<LoupedeckDeviceButton>,
    pub key_grid_dimensions: (usize, usize),
    pub displays: &'static [LoupedeckDeviceDisplayConfiguration]
}

static LOUPEDECK_LIVE_CHARACTERISTIC: LoupedeckDeviceCharacteristics = LoupedeckDeviceCharacteristics {
    vendor_id: 0x2ec2,
    product_id: 0x0004,
    name: "Loupedeck Live",
    available_knobs: enum_set!(LoupedeckDeviceKnob::KnobTopLeft
        | LoupedeckDeviceKnob::KnobCenterLeft
        | LoupedeckDeviceKnob::KnobBottomLeft
        | LoupedeckDeviceKnob::KnobTopRight
        | LoupedeckDeviceKnob::KnobCenterRight
        | LoupedeckDeviceKnob::KnobBottomRight),
    available_buttons: enum_set!(LoupedeckDeviceButton::KnobTopLeft
        | LoupedeckDeviceButton::KnobCenterLeft
        | LoupedeckDeviceButton::KnobBottomLeft
        | LoupedeckDeviceButton::KnobTopRight
        | LoupedeckDeviceButton::KnobCenterRight
        | LoupedeckDeviceButton::KnobBottomRight
        | LoupedeckDeviceButton::N0
        | LoupedeckDeviceButton::N1
        | LoupedeckDeviceButton::N2
        | LoupedeckDeviceButton::N3
        | LoupedeckDeviceButton::N4
        | LoupedeckDeviceButton::N5
        | LoupedeckDeviceButton::N6
        | LoupedeckDeviceButton::N7),
    key_grid_dimensions: (4, 3),
    displays: &[
        LoupedeckDeviceDisplayConfiguration { // Left
            id: 0x4d,
            width: 60,
            height: 270,
            offset: (0, 0),
            endianness: LoupedeckDeviceDisplayEndianness::LittleEndian,
        },
        LoupedeckDeviceDisplayConfiguration { // Center
            id: 0x4d,
            width: 360,
            height: 270,
            offset: (60, 0),
            endianness: LoupedeckDeviceDisplayEndianness::LittleEndian,
        },
        LoupedeckDeviceDisplayConfiguration { // Right
            id: 0x4d,
            width: 60,
            height: 270,
            offset: (420, 0),
            endianness: LoupedeckDeviceDisplayEndianness::LittleEndian,
        },
    ]
};

pub static CHARACTERISTICS: [&LoupedeckDeviceCharacteristics; 1] = [
    &LOUPEDECK_LIVE_CHARACTERISTIC
];