use std::io;
use std::io::{BufReader, Read, Write};
use std::io::ErrorKind::TimedOut;
use std::sync::Arc;
use std::sync::mpsc::{Sender, sync_channel, SyncSender};
use std::thread::spawn;
use bytes::{Buf, BufMut, Bytes, BytesMut};
use serialport::{SerialPort, SerialPortType};
use thiserror::Error;
use crate::characteristics::{CHARACTERISTICS, LoupedeckDeviceCharacteristics};
use crate::discovery::AvailableLoupedeckDevice;

const WS_UPGRADE_REQUEST: &str = r#"GET /index.html
HTTP/1.1
Connection: Upgrade
Upgrade: websocket
Sec-WebSocket-Key: 123abc

"#;

const WS_UPGRADE_RESPONSE_START: &str = "HTTP/1.1 101 Switching Protocols\r\n\
Upgrade: websocket\r\n\
Connection: Upgrade\r\n\
Sec-WebSocket-Accept: ALtlZo9FMEUEQleXJmq++ukUQ1s=";

const MESSAGE_START_BYTE: u8 = 0x82;
const MAX_MESSAGE_LENGTH: usize = u8::MAX as usize;

#[derive(Debug)]
pub struct LoupedeckDevice {
    pub(crate) port: Option<Box<dyn SerialPort>>,
    pub characteristics: &'static LoupedeckDeviceCharacteristics,
}

#[derive(Debug, Error)]
pub enum ConnectError {
    #[error("IO Error: {0}")]
    IO(#[from] io::Error),

    #[error("The device did not respond with the expected handshake response.")]
    WrongHandshakeResponse,
}

impl LoupedeckDevice {
    pub fn new_unchecked(port: Box<dyn SerialPort>, characteristics: &'static LoupedeckDeviceCharacteristics) -> LoupedeckDevice {
        LoupedeckDevice {
            port: Some(port),
            characteristics,
        }
    }

    pub fn discover() -> Result<Vec<AvailableLoupedeckDevice>, serialport::Error> {
        let ports = serialport::available_ports()?;

        Ok(ports.iter().filter_map(|port| {
            if let SerialPortType::UsbPort(info) = &port.port_type {
                let characteristics = CHARACTERISTICS.iter()
                    .find(|c| c.vendor_id == info.vid && c.product_id == info.pid);

                if let Some(characteristics) = characteristics {
                    return Some(AvailableLoupedeckDevice {
                        port_name: port.port_name.clone(),
                        characteristics,
                    });
                }
            }

            None
        }).collect::<Vec<_>>())
    }

    pub fn connect(&mut self) -> Result<(), ConnectError> {
        let mut port = self.port.take().expect("already connected");
        port.write_all(WS_UPGRADE_REQUEST.as_bytes())?;
        port.flush()?;

        let mut buf = [0; WS_UPGRADE_RESPONSE_START.len()];
        port.read_exact(&mut buf)?;

        if buf != WS_UPGRADE_RESPONSE_START.as_bytes() {
            return Err(ConnectError::WrongHandshakeResponse);
        }

        let (sender, receiver) = sync_channel::<Bytes>(8);
        let handle = spawn(move || {
            LoupedeckDevice::handle_messages(port, sender);
        });

        for x in receiver {
            dbg!(&x[1]);
        }

        handle.join().unwrap();

        Ok(())
    }

    fn handle_messages(mut port: Box<dyn SerialPort>, sender: SyncSender<Bytes>) {
        let mut buffer = BytesMut::new();

        loop {
            let mut chunk = BytesMut::zeroed(MAX_MESSAGE_LENGTH);
            let read_result = port.read(&mut chunk);

            if let Err(err) = &read_result {
                if err.kind() == TimedOut {
                    continue
                }
            }

            let len = read_result.unwrap();
            if len == 0 {
                panic!("read 0 bytes");
            }

            chunk.truncate(len);
            buffer.put(chunk);

            loop {
                let start_index = buffer.iter().position(|b| *b == MESSAGE_START_BYTE);

                if let Some(start_index) = start_index {
                    if start_index > 0 {
                        buffer.advance(start_index);

                        if buffer.remaining() == 0 {
                            break;
                        }
                    }

                    let length = buffer[1] as usize + 2;
                    if length > buffer.remaining() {
                        break;
                    } else {
                        let mut message = buffer.split_to(length);
                        message.advance(2);
                        sender.send(message.freeze()).unwrap();
                    }
                } else {
                    break;
                }
            }
        }
    }
}